/*
* Copyright 2023, Charadon
*
* Copying and distribution of this file, with or without modification, are
* permitted in any medium without royalty, provided the copyright notice and
* this notice are preserved. This file is offered as-is, without any
* warranty.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <unistd.h>
#include <windows.h>
#include <winbase.h>
#include <time.h>
#include <errno.h>
#include <ctype.h>
#include "7zip.h"
#include "runtime.h"

// Workaround for there being no realpath on windows.
#define realpath(N,R) _fullpath((R),(N),_MAX_PATH)

void error_at_line(int Line, const char *File, const char *Doing) {
  fprintf(stderr, "Program failed at line %d in %s doing or for reason: %s\n", Line, File, Doing);
  exit(1);
}


char *format_winpath_to_cygpath(const char *Winpath) {
  char Winpath_RW[PATH_MAX];
  if(strncpy(Winpath_RW, Winpath, PATH_MAX-1) == NULL) {
    error_at_line(__LINE__, __FILE__, "strncpy");
  }
  char Cygpath[PATH_MAX];
  // Get the drive letter from the path.
  char DriveLetter = tolower(Winpath_RW[0]);
  if(snprintf(Cygpath, PATH_MAX-1, "/cygdrive/%c", DriveLetter) < 0) {
    error_at_line(__LINE__, __FILE__, "snprintf");
  }
  char PathString[PATH_MAX];
  int a = 0; // For PathString
  for(int i = 2; Winpath_RW[i] != '\0'; i++) {
    switch(Winpath_RW[i]) {
      case '\\':
        PathString[a] = '/';
        break;
      default:
        PathString[a] = Winpath_RW[i];
        break;
    }
    a++;
  }
  PathString[a] = '\0'; // Terminate string.
  if(strncat(Cygpath, PathString, PATH_MAX-1) == NULL) {
    error_at_line(__LINE__, __FILE__, "strncat");
  }
  char *ReturnedPath = NULL;
  ReturnedPath = malloc(sizeof(char)*PATH_MAX);
  if(snprintf(ReturnedPath, PATH_MAX-1, "%s", Cygpath) < 0) {
    error_at_line(__LINE__, __FILE__, "snprintf");
  }
  return ReturnedPath;
}

void dump_file(const char *EmbeddedFile, int EmbeddedFileLength, const char *FilePath, const char *TempDir) {
  FILE *FileToDump = NULL;
  char FilePathRW[PATH_MAX];
  if(strncpy(FilePathRW, FilePath, PATH_MAX-1) == NULL) {
    error_at_line(__LINE__, __FILE__, "strncpy");
  }
  char TempDirRW[PATH_MAX];
  if(strncpy(TempDirRW, TempDir, PATH_MAX-1) == NULL) {
    error_at_line(__LINE__, __FILE__, "strncpy");
  }
  FileToDump = fopen(strncat(TempDirRW, FilePathRW, PATH_MAX-1), "wb");
  if(FileToDump == NULL) {
    error_at_line(__LINE__, __FILE__, "fopen");
  }
  if(fwrite(EmbeddedFile, 1, EmbeddedFileLength, FileToDump) == 0) {
    error_at_line(__LINE__, __FILE__, "fwrite");
  }
  if(fclose(FileToDump) != 0) {
    error_at_line(__LINE__, __FILE__, "fclose");
  }
}

char generate_random_character() {
  char ValidCharacters[] = {
    '1', '2', '3', '4', '5', '6', '7', '8', '9', '0', 'A', 'B', 'C', 'D', 'E', \
    'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', \
    'U', 'V', 'W', 'X', 'Y', 'Z'
  };
  return(ValidCharacters[rand() % sizeof(ValidCharacters)]);
}

// Microsoft sucks.
char *mkdtemp(const char *TempPath) {
  char *TempPathString = NULL;
  TempPathString = malloc(sizeof(char) * PATH_MAX);
  if (strncpy(TempPathString, TempPath, PATH_MAX) == NULL) {
    error_at_line(__LINE__, __FILE__, "strncpy");
  }
  int Length = strlen(TempPathString);
  char *cwd = getcwd(NULL, PATH_MAX);
  if(cwd == NULL) {
    error_at_line(__LINE__, __FILE__, "getcwd");
  }
TryAgain:
  for(int i = 0; i <= 6; i++) {
    TempPathString[Length-i] = generate_random_character();
  }
  // This should fail, as the path shouldn't exist. If it does, reroll.
  int Result = chdir(TempPathString);
  if(Result == 0) {
    if(chdir(cwd) != 0) {
      error_at_line(__LINE__, __FILE__, strerror(errno));
    }
    goto TryAgain;
  }
  if(mkdir(TempPathString) != 0) {
    fprintf(stderr, "Failed to make temporary directory.\n");
    exit(1);
  }
  return TempPathString;
}

int main(int argc, char *argv[]) {
  srand(time(NULL));
  char RealPathLocation[PATH_MAX];
  if(realpath(argv[0], RealPathLocation) == NULL) {
    error_at_line(__LINE__, __FILE__, strerror(errno));
  }
  // Grab the Windows TEMP Environment Variable.
  char *DebugMode = getenv("SH2EXE_DEBUG");
  int Debug = 0;
  if(DebugMode != NULL) {
    if(strncmp(DebugMode, "1", 2048) == 0) {
      Debug = 1;
    }
  }
  char *Temp = getenv("TEMP");
  if(Temp[0] == '\0') {
    error_at_line(__LINE__, __FILE__, "getenv");
  }

  // Create a temporary directory.
  char *Template = "\\sh2exe.XXXXXX";
  char *TempDir = mkdtemp(strncat(Temp, Template, PATH_MAX));
  if(TempDir == NULL) {
    error_at_line(__LINE__, __FILE__, "mkdtemp");
  }

  // Dump embedded files.
  dump_file(SevenZip, sizeof(SevenZip), "\\7zip.exe", TempDir);
  dump_file(RuntimeArchive, sizeof(RuntimeArchive), "\\runtime.7z", TempDir);

  // Extract runtime with 7zip.
  char *CurrentWorkingDirectory = getcwd(NULL, PATH_MAX);
  if(CurrentWorkingDirectory == NULL) {
    error_at_line(__LINE__, __FILE__, "getcwd");
  }
  if (chdir(TempDir) != 0) {
    error_at_line(__LINE__, __FILE__, strerror(errno));
  }
  // We use WinExec since it's the simplest method of launching applications
  int Result = WinExec("\"C:\\Windows\\System32\\cmd.exe\" /C \".\\7zip.exe x .\\runtime.7z > nul && echo \"Done\" > 7ZIP_DONE\"", SW_HIDE);
  if(Result < 31) {
    error_at_line(__LINE__, __FILE__, "Runtime Extraction");
  }
  //Timeout after a minute.
  int Timeout = 0;
  while(access(".\\7ZIP_DONE", F_OK) != 0) {
    if(usleep(10000) != 0) {
      error_at_line(__LINE__, __FILE__, strerror(errno));
    }
    ++Timeout;
    if(Timeout >= 6000) {
      error_at_line(__LINE__, __FILE__, "Timeout reached for runtime extraction.");
    }
  }
  // We don't need runtime.7z anymore, and 7zip can be moved to /usr/bin.
  if(remove(".\\runtime.7z") != 0) {
    error_at_line(__LINE__, __FILE__, strerror(errno));
  }
  if(rename(".\\7zip.exe", ".\\usr\\bin\\7z.exe") != 0) {
    error_at_line(__LINE__, __FILE__, strerror(errno));
  }

  // Fix issue where bash will complain about missing /tmp
  char RuntimeTmpDir[PATH_MAX];
  if(strncpy(RuntimeTmpDir, TempDir, PATH_MAX-1) == NULL) {
    error_at_line(__LINE__, __FILE__, "strncpy");
  }
  if(strncat(RuntimeTmpDir, "\\tmp", PATH_MAX-1) == NULL) {
    error_at_line(__LINE__, __FILE__, "strncat");
  }
  if(mkdir(RuntimeTmpDir) != 0) {
    if(errno != EEXIST) {
      error_at_line(__LINE__, __FILE__, strerror(errno));
    }
  }

  // Set environment variables.
  // PATH
  char *PathEnv = getenv("PATH");
  if(PathEnv == NULL) {
    error_at_line(__LINE__, __FILE__, "getenv");
  }
  char NewPathEnv[USHRT_MAX];
  if(strncpy(NewPathEnv, TempDir, PATH_MAX) == NULL) {
    error_at_line(__LINE__, __FILE__, "strncpy");
  }
  if(strncat(NewPathEnv, "\\usr\\bin;", PATH_MAX) == NULL) {
    error_at_line(__LINE__, __FILE__, "strncat");
  }
  if(strncat(NewPathEnv, PathEnv, USHRT_MAX-1) == NULL) {
    error_at_line(__LINE__, __FILE__, "strncat");
  }
  char PathVariable[USHRT_MAX] = "PATH=";
  if(putenv(strncat(PathVariable, NewPathEnv, USHRT_MAX-1)) != 0) {
    error_at_line(__LINE__, __FILE__, strerror(errno));
  }
  if (chdir(CurrentWorkingDirectory) != 0) {
    error_at_line(__LINE__, __FILE__, strerror(errno));
  }
  // HOME
  char *HomeEnv = getenv("HOMEPATH");
  if(HomeEnv == NULL) {
    error_at_line(__LINE__, __FILE__, "getenv");
  }
  char PosixHomeEnv[USHRT_MAX] = "HOME=";
  if(strncat(PosixHomeEnv, HomeEnv, USHRT_MAX-1) == NULL) {
    error_at_line(__LINE__, __FILE__, "strncat");
  }
  if(putenv(PosixHomeEnv) != 0) {
    error_at_line(__LINE__, __FILE__, strerror(errno));
  }
  // USER
  char *UserEnv = getenv("USERNAME");
  if(UserEnv == NULL) {
    error_at_line(__LINE__, __FILE__, "getenv");
  }
  char PosixUserEnv[USHRT_MAX] = "USER=";
  if(strncat(PosixUserEnv, UserEnv, USHRT_MAX-1) == NULL) {
    error_at_line(__LINE__, __FILE__, "strncat");
  }
  if(putenv(PosixUserEnv) != 0) {
    error_at_line(__LINE__, __FILE__, strerror(errno));
  }
  // EDITOR
  char *EditorEnv = getenv("EDITOR");
  if(EditorEnv == NULL) {
    if(putenv("EDITOR=/usr/bin/nano") != 0) {
      error_at_line(__LINE__, __FILE__, strerror(errno));
    }
  }
  // $SH2EXE_ARGV0
  char ScriptLocationEnv[PATH_MAX] = "SH2EXE_ARGV0=";
  char *ScriptLocationCygpath = format_winpath_to_cygpath(RealPathLocation);
  if(strncat(ScriptLocationEnv, ScriptLocationCygpath, PATH_MAX-1) == NULL) {
    error_at_line(__LINE__, __FILE__, "strncpy");
  }
  if(putenv(ScriptLocationEnv) != 0) {
    error_at_line(__LINE__, __FILE__, strerror(errno));
  }

  // Format command string and then run it.
  char Command[PATH_MAX];
  char CommandArgs[USHRT_MAX];
  for(int i = 1; i < argc; i++) {
    if (strncat(CommandArgs, " ", USHRT_MAX-1) == NULL) {
      error_at_line(__LINE__, __FILE__, "strncat");
    }
    if (strncat(CommandArgs, argv[i], USHRT_MAX-1) == NULL) {
      error_at_line(__LINE__, __FILE__, "strncat");
    }
  }
  char ScriptName[USHRT_MAX] = "sh"; //If SH2EXE_DEBUG is equal to 1, we'll just launch a normal shell.
  if(Debug == 0) {
    if(strncpy(ScriptName, "INSERT_SCRIPT_NAME_HERE", USHRT_MAX) == NULL) {
      error_at_line(__LINE__, __FILE__, "strncpy");
    }
  }
  else {
    if(strncpy(ScriptName, "sh", USHRT_MAX) == NULL) {
      error_at_line(__LINE__, __FILE__, "strncpy");
    }
  }
  // Format command string to run.
  if(snprintf(Command, PATH_MAX-1, "\"%s\\usr\\bin\\sh.exe\" -c \"/usr/bin/%s %s; echo \"$?\" > %s/SCRIPT_DONE\"", TempDir, ScriptName, CommandArgs, format_winpath_to_cygpath(TempDir)) < 0) {
    error_at_line(__LINE__, __FILE__, "snprintf");
  }
  if(WinExec(Command, SW_HIDE) < 31) {
    error_at_line(__LINE__, __FILE__, "Failed to run script.");
  }
  char ScriptDoneFile[PATH_MAX];
  if(snprintf(ScriptDoneFile, PATH_MAX-1, "%s/SCRIPT_DONE", TempDir) < 0) {
    error_at_line(__LINE__, __FILE__, "snprintf");
  }
  // Prevent heavy CPU Usage while waiting for Script to finish running.
  while(access(ScriptDoneFile, F_OK) != 0) {
    usleep(100000);
  }
  FILE *ReturnCodeFile = NULL;
  ReturnCodeFile = fopen(ScriptDoneFile, "r");
  if(ReturnCodeFile == NULL) {
    error_at_line(__LINE__, __FILE__, strerror(errno));
  }
  fscanf(ReturnCodeFile, "%d", &Result);
  if(ReturnCodeFile != NULL) {
    fclose(ReturnCodeFile);
  }
  // Cleanup Temporary Directory.
  char DeleteTempDirPath[PATH_MAX];
  if(strncpy(DeleteTempDirPath, TempDir, PATH_MAX-1) == NULL) {
    error_at_line(__LINE__, __FILE__, "strncpy");
  }
  char RmDirCmd[PATH_MAX] = "\"C:\\Windows\\System32\\cmd.exe\" /C rmdir /Q /S ";
  if(strncat(RmDirCmd, DeleteTempDirPath, PATH_MAX-1) == NULL) {
    error_at_line(__LINE__, __FILE__, "strncat");
  }
  if (WinExec(RmDirCmd, SW_HIDE) < 31) {
    error_at_line(__LINE__, __FILE__, "Couldn't delete temporary directory of runtime.");
  }
  return(Result);
}
